"""Bandwitch, a Python library for computer-aided restriction digests."""

# __all__ = []

from .enzyme_suggestion import (SeparatingDigestionsProblem,
                                IdealDigestionsProblem)
from .migration_analysis import BandsObservation, Clone, ClonesObservations
from .Ladder import Ladder

from .tools import (band_patterns_discrepancy, digestions_list_to_string,
                    load_genbank, list_common_enzymes)
from .bands_predictions import predict_digestion_bands
from .data import enzymes_infos, LADDERS
from .version import __version__




import os
from bandwitch import (SeparatingDigestionsProblem, list_common_enzymes,
                       LADDERS, load_genbank)


# DEFINE THE SEQUENCES AND THE ENZYME SET

# all common enzymes with a site of 6bp and at least 3 known providers
enzymes = list_common_enzymes(site_length=(6,), min_suppliers=3)

records_folder = os.path.join('example_data', 'digestion_selection_data')
sequences = [
    load_genbank(os.path.join(records_folder, f), name=f)
    for f in sorted(os.listdir(records_folder))
]

# DEFINE AND SOLVE THE PROBLEM

problem = SeparatingDigestionsProblem(enzymes=enzymes,
                                      ladder=LADDERS['100_to_4k'],
                                      sequences=sequences,
                                      max_enzymes_per_digestion=1)
score, selected_digestions = problem.select_digestions(max_digestions=2)

# GENERATE A FIGURE OF THE BAND PATTERNS

axes = problem.plot_digestions(
    selected_digestions,
    patterns_props={'label_fontdict': {'rotation': 35}},
    target_file="separating_digestions.png"
)
problem.plot_distances_map(digestions=selected_digestions,
                           target_file="separating_digestions_distances.png")



"""Classes for parsing and analyzis of DNA migration data."""
from collections import OrderedDict, Counter
import matplotlib.pyplot as plt
from io import BytesIO
from matplotlib.backends.backend_pdf import PdfPages

from bandwagon import BandsPattern, BandsPatternsSet
import flametree

try:
    from plateo.exporters.plate_to_matplotlib_plots import PlateColorsPlotter
    from plateo.parsers import plate_from_platemap_spreadsheet
    from plateo.containers import Plate96
    PLATEO_AVAILABLE = True
except:
    PLATEO_AVAILABLE = False


from ..bands_predictions import predict_digestion_bands
from ..plots import plot_cuts_map, plot_all_constructs_cuts_maps
from ..tools import load_genbank
from .Clone import Clone
from .BandsObservation import BandsObservation


class ClonesObservations:
    """All useful informations for a collection of clones to be validated.

    Parameters
    ----------
    clones
      Either a list of Clones (each with a unique name) or a dictionnary
      ``{name: Clone}``.

    constructs_records
      A dictionnary ``{construct_name: biopython_record}`` indicating the
      sequence of the different constructs. For each construct, set the
      attribute ``construct.linear = False`` if the construct is circular.

    """

    def __init__(self, clones, constructs_records):
        """Initialize"""
        if isinstance(clones, (list, tuple)):
            clones = {clone.name: clone for clone in clones}
        self.clones = clones
        self.constructs_records = constructs_records
        self.constructs_digestions = {ct: {} for ct in constructs_records}


    def get_digestion_bands_for_construct(self, construct_id, digestion):
        """Return the bands resulting from the digestion of the construct.

        This function enables some memoization (no digestion is computed
        twice).

        Parameters
        ----------

        construct_id
          ID of the construct as it appears in this object'
          ``constructs_records``.

        digestion
          For instance ``('BamHI', 'XbaI')``

        """
        construct_digestions = self.constructs_digestions[construct_id]
        if digestion not in construct_digestions:
            construct_record = self.constructs_records[construct_id]
            construct_digestions[digestion] = predict_digestion_bands(
                str(construct_record.seq),
                linear=construct_record.__dict__.get('linear', True),
                enzymes=digestion
            )
        return construct_digestions[digestion]

    def get_clone_digestion_bands(self, clone, construct_id=None):
        """Return ``{digestion: bands}`` for all digestions of this clone."""
        if construct_id is None:
            construct_id = clone.construct_id
        return {
            digestion: self.get_digestion_bands_for_construct(
                construct_id=clone.construct_id, digestion=digestion)
            for digestion in clone.digestions
        }

    def validate_all_clones(self, relative_tolerance=0.1,
                            min_band_cutoff=None, max_band_cutoff=None):
        """Return ``{clone: CloneValidation}`` for all clones."""
        return OrderedDict([
            (clone_name, clone.validate_bands(
                bands_by_digestion=self.get_clone_digestion_bands(clone),
                relative_tolerance=relative_tolerance,
                min_band_cutoff=min_band_cutoff,
                max_band_cutoff=max_band_cutoff
            ))
            for clone_name, clone in self.clones.items()
        ])

    def identify_all_clones(self, relative_tolerance=0.1,
                            min_band_cutoff=None, max_band_cutoff=None):
        """Return ``{clone: {construct_id: CloneValidation}}`` for all clones.

        Parameters
        ----------
        relative_tolerance
          Tolerance, as a ratio of the full ladder span. If =0.1, then the
          discrepancy will have a value of 1 when a band's nearest
          correspondent in the other pattern is more that 10% of the ladder
          span apart.

        min_band_cutoff
          Discrepancies involving at least one band below this minimal band
          size will be ignored. By default, it will be set to the smallest
          band size in the ladder.

        max_band_cutoff
          Discrepancies involving at least one band above this minimal band
          size will be ignored. By default, it will be set to the smallest
          band size in the ladder.

        """
        return OrderedDict([
            (clone_name, {
                construct_id: clone.validate_bands(
                    bands_by_digestion=self.get_clone_digestion_bands(
                        clone, construct_id=construct_id),
                    relative_tolerance=relative_tolerance,
                    min_band_cutoff=min_band_cutoff,
                    max_band_cutoff=max_band_cutoff
                )
                for construct_id in self.constructs_records
            })
            for clone_name, clone in self.clones.items()
        ])

    def validations_summary(self, validations, sort_clones_by_score=True):
        """Return ``{construct_id: [CloneValidation, ...]}``.

        To each construct corresponds a list of the validation of all clones
        associated with that construct, from the best-scoring to the
        least-scoring.
        """
        results = OrderedDict()
        for clone_name, validation in validations.items():
            construct_id = self.clones[clone_name].construct_id
            if construct_id not in results:
                results[construct_id] = []
            results[construct_id].append(validation)
        if sort_clones_by_score:
            for cst_id, validations in results.items():
                results[cst_id] = sorted(validations,
                                         key=lambda v: v.max_discrepancy)
        return results

    def plot_validations_plate_map(self, validations, target=None,
                                   ax=None):
        """Plot a map of the plate with passing/failing wells in green/red."""

        if not PLATEO_AVAILABLE:
            raise ImportError('This function requires Plateo installed')
        plate = Plate96('Validations map')
        constructs_colors = {}
        for well in plate.iter_wells():
            if well.name not in validations:
                well.data.bg_color = None
                well.data.pass_color = (1, 1, 1, 0.1)
            else:
                validation = validations[well.name]
                construct = validation.clone.construct_id
                if construct not in constructs_colors:
                    alpha = ((0.35 * len(constructs_colors) % 1) + .2) / 2
                    constructs_colors[construct] = (.4, .4, .76, alpha)
                well.data.bg_color = constructs_colors[construct]
                well.data.pass_color = ((.43, .92, .49) if validation.passes
                                        else (.91, 0.43, 0.43))
        bg_plotter = PlateColorsPlotter(lambda w: w.data.bg_color)
        ax, _ = bg_plotter.plot_plate(plate, figsize=(10, 5), ax=ax)
        pass_plotter = PlateColorsPlotter(lambda w: w.data.pass_color,
                                          well_radius=250)
        pass_plotter.plot_plate(plate, ax=ax)
        if target is not None:
            ax.figure.savefig(target, bbox_inches='tight')
            plt.close(ax.figure)
        else:
            return ax

    def plot_all_validations_patterns(self, validations, target=None,
                                      per_digestion_discrepancy=False):
        """Plot a Graphic report of the gel validation.

        The report displays the patterns, with green and red backgrounds
        depending on whether they passed the validation test.

        Parameters
        ----------

        target_file
          File object or file path where to save the figure. If provided, the
          function returns none, else the function returns the axes array of
          the final figure

        relative_tolerance
          Relative error tolerated on each band for the ovserved patterns to
          be considered similar to the expected patterns.

        min_band_cutoff
          Bands with a size below this value will not be considered

        max_band_cutoff
          Bands with a size above this value will not be considered

        """

        summary = self.validations_summary(validations)
        max_x = Counter([
            clone.construct_id
            for clone in self.clones.values()
        ]).most_common(1)[0][1] + 2
        digestions_by_construct = {construct: [] for construct in summary}
        for construct, validations in summary.items():
            for validation in validations:
                for digestion in validation.discrepancies:
                    if digestion not in digestions_by_construct[construct]:
                        digestions_by_construct[construct].append(digestion)
        ladder = list(validations[0].clone.digestions.values())[0].ladder
        pdf_io = BytesIO()
        with PdfPages(pdf_io) as pdf:
            for construct_id, validations in summary.items():
                digestions = digestions_by_construct[construct_id]
                band_patterns = OrderedDict()
                for i, digestion in enumerate(digestions):
                    reference_bands = self.get_digestion_bands_for_construct(
                        construct_id, digestion)
                    reference = BandsPattern(
                        bands=reference_bands,
                        ladder=ladder,
                        label="exp." if (i == 0) else None,
                        background_color="#c6dcff",
                        corner_note="Total: %d bp" % sum(reference_bands),
                        global_bands_props={"label_fontdict": {"size": 5}}
                    )
                    sorted_bands = sorted(reference.bands,
                                          key=lambda b: -b.dna_size)
                    for band_name, band in zip("abcdefghijklm", sorted_bands):
                        band.label = band_name
                    patterns = [
                        validation.to_bandwagon_bandpattern(
                            digestion=digestion,
                            per_digestion_discrepancy=per_digestion_discrepancy,
                            label='auto' if (i == 0) else None
                        )
                        for validation in validations
                    ]
                    band_patterns[digestion] = BandsPatternsSet(
                        [reference] + patterns, ladder=ladder,
                        label=" + ".join(digestion),
                        ladder_ticks=5,
                        global_patterns_props={
                            "label_fontdict": {"rotation": 60}}
                    )

                digestions = digestions_by_construct[construct_id]
                fig, axes = plt.subplots(
                    len(digestions), 1, sharex=True,
                    figsize=(1.1 * max_x, 3 * len(digestions)),
                )
                if len(digestions) == 1:
                    axes = [axes]
                axes[-1].set_xlim(xmax=max_x)

                for ax, (dig, pattern_set) in zip(axes, band_patterns.items()):
                    pattern_set.plot(ax)

                axes[-1].set_xlabel(construct_id,
                                    fontdict=dict(size=14, weight='bold'),
                                    ha='left', position=(0.0, 0.1))
                pdf.attach_note(construct_id)
                pdf.savefig(fig, bbox_inches="tight")
                plt.close(fig)
        pdf_data = pdf_io.getvalue()
        if target is None:
            return pdf_data
        elif target == 'base64':
            return 'data:application/pdf;base64,' + pdf_data.decode("utf-8")
        else:
            with open(target, 'wb') as f:
                f.write(pdf_data)

    def write_identification_report(self, target_file=None,
                                    relative_tolerance=0.1,
                                    min_band_cutoff=None,
                                    max_band_cutoff=None,
                                    plot_constructs_cuts=False):
        """Plot a Graphic report of the gel validation.

        The report displays the patterns, with green and red backgrounds
        depending on whether they passed the validation test.

        Parameters
        ----------

        target_file
          File object or file path where to save the figure. If provided, the
          function returns none, else the function returns the axes array of
          the final figure

        relative_tolerance
          Relative error tolerated on each band for the ovserved patterns to
          be considered similar to the expected patterns.

        min_band_cutoff
          Bands with a size below this value will not be considered

        max_band_cutoff
          Bands with a size above this value will not be considered

        """
        bands_validities = self.compute_all_bands_validities(
            relative_tolerance=relative_tolerance,
            min_band_cutoff=min_band_cutoff,
            max_band_cutoff=max_band_cutoff)
        L = len(self.constructs_records)
        max_x = max(
            len(measures)
            for measures in self.observed_bands.values()
        ) + 1
        fig, axes = plt.subplots(L, 2, figsize=(2.2 * max_x, 3 * L))
        axes_validities = zip(axes, bands_validities.items())
        for (ax1, ax2), (construct_id, validities) in axes_validities:
            reference = BandsPattern(
                self.expected_bands[construct_id], ladder=self.ladder,
                label="exp.", background_color="#c6dcff",
                corner_note="Total: %d bp" % sum(
                    self.expected_bands[construct_id]),
                global_bands_props={"label_fontdict": {"size": 5}}
            )
            sorted_bands = sorted(reference.bands, key=lambda b: -b.dna_size)
            for band_name, band in zip("abcdefghijklm", sorted_bands):
                band.label = band_name
            patterns = [

                BandsPattern(
                    self.observed_bands[construct_id][measure_name],
                    corner_note="Total: %d bp" % sum(
                        self.observed_bands[construct_id][measure_name]),
                    ladder=self.ladder,
                    label=measure_name,
                    gel_image=self.migration_images[measure_name],
                    background_color="#aaffaa" if
                                     validities[measure_name]
                                     else "#ffaaaa"
                )
                for measure_name in self.observed_bands[construct_id]
            ]
            patterns_set = BandsPatternsSet(
                [reference] + patterns, ladder=self.ladder,
                label=construct_id, ladder_ticks=5,
                global_patterns_props={"label_fontdict": {"rotation": 60}}
            )
            ax1.set_xlim(0.5, max_x + 2)
            record = self.constructs_records[construct_id]
            plot_cuts_map(record, enzymes=self.digestions[construct_id],
                          ax=ax2)
            patterns_set.plot(ax1)
        fig.subplots_adjust(hspace=0.3)
        if target_file is not None:
            fig.savefig(target_file, bbox_inches="tight")
            plt.close(fig)
        else:
            return axes

    def plot_all_constructs_cuts_maps(self, target=None, figsize=(12, 4)):
        """Plot schemas of all constructs with cuts, in a multipage PDF.

        Parameters
        ----------
        target
          Either None (at which case the function returns raw data of the PDF)
          or 'base64' (the function returns base64-encoded data, e.g. for
          web transfer), or a file path to be written to
        figsize
          The size in inches of each figure (=page of the pdf).

        """
        constructs_digestions = OrderedDict([
            (construct_id, set())
            for construct_id in self.constructs_records
        ])
        for clone in self.clones.values():
            for digestion in clone.digestions:
                constructs_digestions[clone.construct_id].add(digestion)
        return plot_all_constructs_cuts_maps([
            (self.constructs_records[construct_id], digestion)
            for construct_id, digestions in constructs_digestions.items()
            for digestion in sorted(digestions)
        ], target=target, figsize=figsize)

    def from_files(records_path, constructs_map_path, aati_zip_path,
                   digestions_map_path=None, digestion=None):
        constructs_records = {
            f._name_no_extension: load_genbank(
                f.open('r'), linear=False, name=f._name_no_extension)
            for f in flametree.file_tree(records_path)._all_files
        }
        constructs_records = OrderedDict(sorted(constructs_records.items()))

        constructs_plate = plate_from_platemap_spreadsheet(
            constructs_map_path, data_field='construct', headers=True)
        constructs_map = OrderedDict([
            (well.name, well.data.construct)
            for well in constructs_plate.iter_wells(direction='row')
            if 'construct' in well.data
            and str(well.data.construct) != 'nan'
        ])

        if digestion:
            digestion = tuple(digestion)
            digestions_map = OrderedDict([
                (wellname, digestion)
                for wellname in constructs_map
            ])
        else:
            digestions_plate = plate_from_platemap_spreadsheet(
                digestions_map_path, data_field='digestion', headers=True)
            digestions_map = OrderedDict([
                (well.name, tuple(well.data.digestion.split(', ')))
                for well in digestions_plate.iter_wells(direction='row')
                if 'digestion' in well.data
                and str(well.data.digestion) != 'nan'
            ])

        observations = BandsObservation.from_aati_fa_archive(aati_zip_path)
        clones = Clone.from_bands_observations(observations, constructs_map,
                                               digestions_map)
        clones_observations = ClonesObservations(clones, constructs_records)
        return clones_observations



from bandwagon.ladders import ladder_from_aati_fa_calibration_table
from bandwagon import BandsPattern
from ..tools import band_patterns_discrepancy
try:
    from plateo.parsers import plate_from_aati_fragment_analyzer_zip
    PLATEO_AVAILABLE = True
except:
    PLATEO_AVAILABLE = False

class BandsObservation:
    """One observation of a bands pattern.

    Parameters
    ----------
    name
      Name of the observation (used for the top label in plots)

    bands
      A BandPattern object

    ladder
      A BandPattern object representing a ladder.

    migration_image
      Optional RGB array (HxWx3) representing the gel "image" (which will be
      displayed on the side of the plot).

    """

    def __init__(self, name, bands, ladder, migration_image=None):
        """Initialize."""
        self.name = name
        self.bands = sorted(bands)
        self.ladder = ladder
        self.migration_image = migration_image

    @staticmethod
    def from_aati_fa_archive(archive_path, min_rfu_size_ratio=0.3):
        """Return a dictionnary of all band observations in AATI output files.

        Parameters
        ----------
        archive_path
          A path to a ZIP file containing all files output by the AATI fragment
          analyzer.

        min_rfu_size_ratio
          Cut-off ratio to filter-out bands whose intensity is below some
          threshold. The higher the value, the more bands will be filtered out.

        Returns
        -------
        A dictionary ``{'A1': BandsObservation(), 'A2': ...}`` containing the
        measured pattern information for a whole 96-well microplate.

        """
        if not PLATEO_AVAILABLE:
            raise ImportError("Plateo must be installed to parse AATI zips.")

        plate = plate_from_aati_fragment_analyzer_zip(archive_path)
        ladder_data = plate.data.ladder
        ladder = ladder_from_aati_fa_calibration_table(dataframe=ladder_data)
        # get the set of constructs sorted by order of columnwise appearance:

        def band_is_strong_enough(band):
            """Return True iff the band's intensity is above the set level."""
            return (1.0 * band["RFU"] / band["Size (bp)"] > min_rfu_size_ratio)

        return {
            well.name: BandsObservation(
                name=well.name,
                ladder=ladder,
                bands=[
                    band["Size (bp)"]
                    for band in well.data.bands.values()
                    if band_is_strong_enough(band)
                ],
                migration_image=well.data.migration_image
            )
            for well in plate.iter_wells()
        }

    def patterns_discrepancy(self, other_bands, relative_tolerance=0.1,
                             min_band_cutoff=None, max_band_cutoff=None):
        """Return the maximal discrepancy between two band patterns.

        The discrepancy is defined as the largest distance between a band in
        one pattern and the closest band in the other pattern.

        Parameters
        ----------
        other_bands
          A list of bands (integers) to be compared with the current bands

        relative_tolerance
          Tolerance, as a ratio of the full ladder span. If =0.1, then the
          discrepancy will have a value of 1 when a band's nearest
          correspondent in the other pattern is more that 10% of the ladder
          span apart.

        min_band_cutoff
          Discrepancies involving at least one band below this minimal band
          size will be ignored. By default, it will be set to the smallest
          band size in the ladder.

        max_band_cutoff
          Discrepancies involving at least one band above this minimal band
          size will be ignored. By default, it will be set to the smallest
          band size in the ladder.

        """
        ladder_min, ladder_max = self.ladder.dna_size_span
        if min_band_cutoff is None:
            min_band_cutoff = ladder_min
        if max_band_cutoff is None:
            max_band_cutoff = ladder_max
        return band_patterns_discrepancy(
            other_bands, self.bands, ladder=self.ladder,
            relative_tolerance=relative_tolerance,
            zone=[min_band_cutoff, max_band_cutoff],
            reference_and_gel=True
        )

    def to_bandwagon_bandpattern(self, background_color=None, label='auto'):
        """Return a pattern version for the plotting library Bandwagon.

        If label is left to 'auto', it will be the pattern's name.
        """
        if label == 'auto':
            label = self.name
        return BandsPattern(
            self.bands,
            corner_note="Total: %d bp." % sum(self.bands),
            ladder=self.ladder,
            label=label,
            gel_image=self.migration_image,
            background_color=background_color
        )



from collections import OrderedDict
from .CloneValidation import CloneValidation


class Clone:
    """Gather all informations necessary to validate a clone.

    Parameters
    ----------
    name
      Name of the clone. Could be for instance a microplate well.

    digestions
      A dictionnary ``{digestion: CloneObservation}`` where ``digestion`` is
      of the form ``('EcoRI', 'BamHI')``.

    construct_id
      ID of the construct to be validated. This is used to group clones
      by construct in the validation reports.
    """

    def __init__(self, name, digestions, construct_id=None):
        """Initialize."""
        self.name = name
        self.digestions = digestions
        self.construct_id = construct_id

    def validate_bands(self, bands_by_digestion, relative_tolerance=0.1,
                       min_band_cutoff=None, max_band_cutoff=None):
        """Return a validation results (comparison of observed and expected).

        The result is a CloneValidation object.

        Parameters
        ----------
        bands_by_digestion
          A dictionnary ``{digestion: [bands]}`` where ``digestion`` is
          of the form ``('EcoRI', 'BamHI')``, and bands is a list of band sizes

        relative_tolerance
          Tolerance, as a ratio of the full ladder span. If =0.1, then the
          discrepancy will have a value of 1 when a band's nearest
          correspondent in the other pattern is more that 10% of the ladder
          span apart.

        min_band_cutoff
          Discrepancies involving at least one band below this minimal band
          size will be ignored. By default, it will be set to the smallest
          band size in the ladder.

        max_band_cutoff
          Discrepancies involving at least one band above this minimal band
          size will be ignored. By default, it will be set to the smallest
          band size in the ladder.

        """
        for digestion, observation in self.digestions.items():
            discrepancies = {
                digestion: observation.patterns_discrepancy(
                    other_bands=bands_by_digestion[digestion],
                    relative_tolerance=relative_tolerance,
                    min_band_cutoff=min_band_cutoff,
                    max_band_cutoff=max_band_cutoff
                )
                for digestion, observation in self.digestions.items()
            }
        return CloneValidation(self, bands_by_digestion,
                               discrepancies=discrepancies)

    @staticmethod
    def from_bands_observations(observations, constructs_map, digestions_map,
                                clones_map=None):
        if isinstance(observations, (dict, OrderedDict)):
            observations = list(observations.values())
        if clones_map is None:
            clones_map = {name: name for name in constructs_map}
        clones = OrderedDict()

        for obs in observations:
            if obs.name not in clones_map:
                continue
            clone_name = clones_map[obs.name]
            if clone_name not in clones:
                clones[clone_name] = Clone(
                    name=clone_name,
                    construct_id=constructs_map.get(clone_name, 'no_clone'),
                    digestions={}
                )
            digestion = digestions_map[obs.name]
            clones[clone_name].digestions[digestion] = obs

        return clones



from bandwitch import IdealDigestionsProblem, LADDERS, load_genbank
import os

# DEFINE THE CRITERIA FOR ACCEPTABLE BAND PATTERNS

class MyIdealDigestionsProblem(IdealDigestionsProblem):
    def migration_score(self, migration):
        """Are there 2-3 bands between 30% and 70% of the migration span?"""
        min_migration = 0.7 * self.migration_min + 0.3 * self.migration_max
        max_migration = 0.3 * self.migration_min + 0.7 * self.migration_max
        bands_in_central_zone = [band for band in migration
                                 if min_migration <= band <= max_migration]
        return 2 <= len(bands_in_central_zone) <= 3

# DEFINE THE SEQUENCES AND THE ENZYME SET

enzymes = ["EcoRI", "BamHI", "XhoI", "EcoRV", "SpeI", "XbaI", "NotI",
           "SacI", "SmaI", "HindIII", "PstI"]

records_folder = os.path.join('example_data', 'digestion_selection_data')
sequences = [
    load_genbank(os.path.join(records_folder, f), name=f)
    for f in sorted(os.listdir(records_folder))
]

# DEFINE AND SOLVE THE PROBLEM

problem = MyIdealDigestionsProblem(enzymes=enzymes,
                                   ladder=LADDERS['100_to_4k'],
                                   sequences=sequences,
                                   max_enzymes_per_digestion=2)
score, selected_digestions = problem.select_digestions(max_digestions=1)

# GENERATE A FIGURE OF THE BAND PATTERNS

axes = problem.plot_digestions(
    digestions=selected_digestions,
    patterns_props={'label_fontdict': {'rotation': 35}}
)
axes[0].figure.savefig("custom_ideal_digestions.png", bbox_inches="tight")



from bandwitch import IdealDigestionsProblem, LADDERS, load_genbank
import os

# DEFINE THE SEQUENCES AND THE ENZYME SET

enzymes = ["EcoRI", "BamHI", "XhoI", "EcoRV", "SpeI", "XbaI",
           "NotI", "SacI", "SmaI", "HindIII", "PstI"]

records_folder = os.path.join('example_data', 'digestion_selection_data')
sequences = [
    load_genbank(os.path.join(records_folder, f), name=f)
    for f in sorted(os.listdir(records_folder))
]

# DEFINE AND SOLVE THE PROBLEM
problem = IdealDigestionsProblem(enzymes=enzymes,
                                 ladder=LADDERS['100_to_4k'],
                                 sequences=sequences,
                                 max_enzymes_per_digestion=2)

# Here we ask for one digestion
score, selected_digestions = problem.select_digestions(max_digestions=1)

# GENERATE A FIGURE OF THE BAND PATTERNS

problem.plot_digestions(
  digestions=selected_digestions,
  patterns_props={'label_fontdict': {'rotation': 35}}
  target_file="ideal_digestions.png"
)



import os
from ..Ladder import Ladder

this_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)))
ladders_path = os.path.join(this_dir, 'ladders_data')
infos_file = 'ladders_infos.csv'
with open(os.path.join(ladders_path, infos_file), 'r') as f:
    infos = dict(line.split(';') for line in f.read().split('\n')[1:])


class LadderDict(dict):
    pass

LADDERS = LadderDict()
for filename in os.listdir(ladders_path):
    if filename == infos_file:
        continue
    ladder_name = filename.split('.')[0]
    with open(os.path.join(ladders_path, filename), 'r') as f:
        ladder = Ladder(
            infos=infos[ladder_name],
            name=ladder_name, bands=dict(
                tuple(int(e) for e in line.split(';'))
                for line in f.read().split('\n')[1:]
            )
        )
    LADDERS[ladder_name] = LADDERS.__dict__['ladder_' + ladder_name] = ladder



from collections import OrderedDict
import itertools

import numpy as np

from .DigestionProblem import DigestionProblem
from ..tools import max_min_distance

try:
    import matplotlib.pyplot as plt
    PLOTS_AVAILABLE = True
except ImportError:
    PLOTS_AVAILABLE = False

class SeparatingDigestionsProblem(DigestionProblem):
    """Problem: find best digestion(s) to identify constructs.

    Provided a set of constructs (possibly from a combinatorial assembly),
    find a list of digestions such that all pair of constructs have different
    migration patterns for at least one digestion of the list.

    """

    """General class for solving digestions problems.

    By digestion problem we mean *enzymatic* digestion problems. For other
    kinds of digestion problems, consult your proctologist.

    Other digestion problems subclass this problem and implement a computation
    of coverages which depends on the problem.

    Parameters
    ----------

    sequences
      An (ordered) dictionary of the form {sequence_name: sequence} where the
      sequence is an ATGC string

    enzymes
      List of the names of the enzymes to consider, e.g. ``['EcoRI', 'XbaI']``.

    ladder
      A Ladder object representing the ladder used for migrations.

    linear
      True for linear sequences, false for circular sequences

    max_enzymes_per_digestion
      Maximal number of enzymes that can go in a single digestion.
      Experimentally the best is 1, but you can try 2, or 3 in desperate
      situations. Not sure if more enzymes will work.

    relative_error
      Variance of the bands measured during the migration, given as a
      proportion of the total migration span (difference between the migration
      of the ladder's smallest and largest bands).
    """

    def __init__(self, enzymes, ladder, sequences=None, categories=None,
                 linear=False, max_enzymes_per_digestion=1,
                 min_discrepancy='auto',
                 relative_migration_precision=0.1):
        """Initialize."""
        if isinstance(sequences, (list, tuple)):
            sequences = OrderedDict([(r.id, str(r.seq)) for r in sequences])

        if categories is None:
            categories = OrderedDict([
                (seq_name, {seq_name: sequence})
                for seq_name, sequence in sequences.items()
            ])
        else:
            sequences = OrderedDict([
                (seq_name, sequence)
                for category, seqs in categories.items()
                for seq_name, sequence in seqs.items()
            ])
        self.categories = categories

        DigestionProblem.__init__(
            self, sequences=sequences, enzymes=enzymes, ladder=ladder,
            linear=linear, max_enzymes_per_digestion=max_enzymes_per_digestion,
            relative_migration_precision=relative_migration_precision)

    def _compute_elements(self):
        category_pairs = itertools.combinations(self.categories.values(), 2)
        return set(
            (seq1, seq2)
            for category1, category2 in category_pairs
            for seq1 in category1
            for seq2 in category2
        )

    def _parameter_element_score(self, digestion, sequences_pair):
        """See max_patterns_difference."""
        sequence1, sequence2 = sequences_pair
        digestion1, digestion2 = [self.sequences_digestions[s][digestion]
                                  for s in (sequence1, sequence2)]
        if ((digestion1['same_as'] == digestion2['same_as'])
                and digestion1['same_as'] in self.scores[sequences_pair]):
            return self.scores[sequences_pair][digestion1['same_as']]
        #     () and
        # same_as = tuple(sorted(digestion1['same_as'], digestion2['same_as']))

        mini, maxi = self.migration_min, self.migration_max
        margin = self.relative_migration_precision * self.migration_span / 2.0
        distance = max_min_distance(digestion1['migration'],
                                    digestion2['migration'],
                                    zone=[mini + margin, maxi - margin])
        return 1.0 * distance / self.migration_span

    @staticmethod
    def _score_to_color(score, maxi=0.1):
        """Transform a similarity score to a green/red color.

        Parameters
        ----------
        score
          Value between 0 (perfect similarity, green) and 1 (red)

        maxi
          Value of the score above which everything appears completely red.
          Below this value the color goes progressively from red to green in 0.

        """
        return (max(0, min(1, score / maxi)),
                min(1, max(0, 1 - score / maxi)), 0, .5)

    def plot_distances_map(self, digestions, ax=None, target_file=None):
        """Plot how well the digestions separate each construct pair.

        Parameters
        ----------
        digestions
          A list of digestions, eg ``[('EcoRV'), ('XbaI', 'MfeI')]``.

        ax
          A matplotlib ax on which to plot, if none is provided, one is created
          and returned at the end.

        target_file
          The name of the (PNG, SVG, JPEG...) file in which to write the plot.

        Returns
        -------

        axes
          The axes of the generated figure (if a target file is written to,
          the figure is closed and None is returned instead)

        """

        if not PLOTS_AVAILABLE:
            raise ImportError("Plots require Matplotlib/Bandwagon installed.")
        grid = np.zeros(2 * (len(self.sequences),))
        for i, s1 in enumerate(self.sequences):
            for j, s2 in enumerate(self.sequences):
                if i >= j:
                    grid[i, j] = np.nan
                else:
                    scores = [
                        self._parameter_element_score(digestion, (s1, s2))
                        for digestion in digestions
                    ]
                    grid[i, j] = max(scores)
        if ax is None:
            _, ax = plt.subplots(1, figsize=2 * (0.8 * len(grid),))
        ax.imshow(grid[:, ::-1], interpolation='nearest', cmap='OrRd_r',
                  vmin=0, vmax=self.relative_migration_precision)
        for i in range(len(grid)):
            for j in range(len(grid)):
                if i > j:
                    ax.text(
                        len(self.sequences) - i - 1, j,
                        "%d%%" % (100 * grid[j, i]),
                        fontdict=dict(color='black', weight='bold', size=14),
                        horizontalalignment='center',
                        verticalalignment='center'
                    )

        ax.set_yticks(range(len(grid)))
        ax.set_yticklabels(list(self.sequences)[:-1], size=14,
                           fontdict={'weight': 'bold'})
        ax.set_xticks(range(len(grid)))
        ax.xaxis.set_ticks_position('top')
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.set_xticklabels([" " + s for s in list(self.sequences)[1:][::-1]],
                           rotation=90, size=14, fontdict={'weight': 'bold'})
        ax.set_xlim(-0.5, len(self.sequences) - 1.5)
        ax.set_ylim(len(self.sequences) - 1.5, -0.5)

        if target_file is not None:
            ax.figure.savefig(target_file, bbox_inches='tight')
            plt.close(ax.figure)
        return ax



from lxml import html
import requests
from collections import OrderedDict, Counter

if __name__ == "__main__":
    website_url = "http://rebase.neb.com"

    # GET THE CONTENT FROM ALL PROVIDER PAGES
    all_providers_page = requests.get(website_url + "/cgi-bin/lapsupplist")
    parsed_providers_page = html.fromstring(all_providers_page.text)
    providers_pages = OrderedDict(sorted([
        (e.text, requests.get((website_url + e.values()[0])))
        for e in parsed_providers_page.xpath('//a')
        if e.values()[0].startswith("/cgi-bin/laplist")
    ]))

    # CREATE A TABLE WITH ALL ENZYMES FROM ALL PAGES
    full_table = []
    for provider, page in providers_pages.items():
        parsed_page = html.fromstring(page.text)
        tables = parsed_page.xpath('//table[@border=2]')
        for table in tables:
            parsed_table = [
                [v.text_content() for v in row]
                for row in table[2:]
            ]
            if (len(parsed_table) == 0) or (parsed_table[0][0] != 'overlaps:'):
                continue
            header, parsed_table = parsed_table[0], parsed_table[1:]
            full_table += [tuple(row) for row in parsed_table]

    # REMOVE (AND COUNT) DUPLICATES BETWEEN PROVIDERS. WRITE TO A CSV
    counter = Counter(full_table)
    enzymes = [k[0] for k in counter.keys()]
    assert len(enzymes) == len(set(enzymes))
    with open("enzymes_data.csv", "w+") as f:
        f.write("\n".join(
            [";".join(header + ['suppliers'])] + sorted([
                ";".join(line + (str(n),))
                for line, n in counter.items()
            ])))



from collections import OrderedDict
from io import BytesIO

from Bio import Restriction
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from dna_features_viewer import (GraphicRecord, GraphicFeature,
                                 BiopythonTranslator)

def create_cuts_map_graphic_record(record, enzymes):
    """Create a DnaFeaturesViewer graphic record showing the cuts.

    Parameters
    ----------
    record
      A Biopython record

    enzymes
      A list of enzyme names. e.g. ('EcoRI'. 'BamHI')

    """
    batch = Restriction.RestrictionBatch(enzymes)
    cuts_dict = batch.search(record.seq)
    all_cuts = sorted(
        set([0, len(record)] +
            [c for cc in cuts_dict.values() for c in cc])
    )
    bands = list(zip(all_cuts, all_cuts[1:]))
    if not record.__dict__.get('linear', True) and (len(bands) > 1):
        start, end = bands.pop()
        band0 = [-(end - start), bands[0][1]]
        if bands == []:
            bands = [band0]
        else:
            bands[0] = band0
    sorted_bands = sorted(
        bands, key=lambda start_end: start_end[0] - start_end[1])
    gr_cuts = GraphicRecord(len(record), [
        GraphicFeature(_start, _end,  strand=1, label=name,
                       color="#ede15c", thickness=5)
        for name, (_start, _end) in zip("abcdefghijklmn", sorted_bands)
    ])
    gr_cuts.split_overflowing_features_circularly()
    return gr_cuts, all_cuts

def plot_cuts_map(record, enzymes, ax):
    """Plot the record and its cuts one above the other.

    Parameters
    ----------
    record
      A Biopython record

    enzymes
      A list of enzyme names. e.g. ('EcoRI'. 'BamHI')

    ax
      Matplotlib ax on which to plot the figure.

    """

    gr_cuts, all_cuts = create_cuts_map_graphic_record(
        record, enzymes)
    gr_cuts.plot(ax, fontsize=7, with_ruler=False)
    for cut in all_cuts:
        ax.axvline(cut, ls=":", color="k", lw=0.5)

    def features_prop(f):
        return dict(label=f.qualifiers.get("source", [False])[0],
                    color="#93a8ea",
                    thickness=5)

    def features_filter(f):
        return features_prop(f)["label"]

    translator = BiopythonTranslator([features_filter], features_prop)
    gr_record = translator.translate_record(record)
    gr_record.plot(ax, fontsize=4, level_offset=7)

def plot_all_constructs_cuts_maps(record_digestion_pairs, target=None,
                                  figsize=(12, 4)):
    """Plot schemas of all constructs with cuts, in a multipage PDF.

    Parameters
    ----------
    target
      Either None (at which case the function returns raw data of the PDF)
      or 'base64' (the function returns base64-encoded data, e.g. for
      web transfer), or a file path to be written to
    figsize
      The size in inches of each figure (=page of the pdf).

    """
    pdf_io = BytesIO()
    with PdfPages(pdf_io) as pdf:
        for record, digestion in record_digestion_pairs:
            fig, ax = plt.subplots(1, figsize=figsize)
            plot_cuts_map(record, digestion, ax)
            title = record.id + '\n' + " + ".join(digestion)
            ax.set_title(title, fontdict=dict(weight='bold'))
            pdf.attach_note(record.id)
            pdf.savefig(fig, bbox_inches="tight")
            plt.close(fig)
    pdf_data = pdf_io.getvalue()
    if target is None:
        return pdf_data
    elif target == 'base64':
        return 'data:application/pdf;base64,' + pdf_data.decode("utf-8")
    else:
        with open(target, 'wb') as f:
            f.write(pdf_data)



import ez_setup
ez_setup.use_setuptools()

from setuptools import setup, find_packages

exec(open('bandwitch/version.py').read()) # loads __version__

setup(
    name='bandwitch',
    version=__version__,
    author='Zulko',
    description='Enzyme selection for DNA verification and identification',
    long_description=open('pypi-readme.rst').read(),
    license='MIT',
    keywords="Restriction enzyme synthetic biology DNA band patterns",
    packages= find_packages(exclude='docs'),
    include_package_data=True,
    install_requires=('tqdm', 'biopython', 'scipy', 'matplotlib',
                      'bandwagon', 'dna_features_viewer', 'flametree')
)



import os

this_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)))
csv_path = os.path.join(this_dir, "enzymes_data", "enzymes_infos.csv")

with open(csv_path, "r") as f:
    _lines = f.read().split("\n")
    _fields = _lines[0].split(";")
    _replacements = dict([("N/A", False), ("+", True), ("-", True)] +
                         [(str(i), i) for i in range(50)])
    enzymes_infos = {
        _line.split(";")[0]: dict(zip(_fields, [
            _replacements.get(e, e) for e in _line.split(";")]))
        for _line in _lines[1:]
    }



"""Module for digestion pattern prediction."""

from Bio import Restriction
from Bio.Seq import Seq
from collections import OrderedDict


def _compute_bands_from_cuts(cuts, sequence_length, linear=True):
    """Compute the size of the obtained bands from the position of cuts.

    Returns a list of band sizes.

    Parameters
    ----------
    cuts
      Location of the different cuts on the plasmid.

    sequence_length
      Length of the DNA molecule

    linear
      True for a linear DNA molecule, False for a circular DNA molecule.

    """
    cuts = [0] + sorted(list(set(cuts))) + [sequence_length]
    bands = [cut2 - cut1 for cut1, cut2 in zip(cuts, cuts[1:])]
    if not linear and len(bands) > 1:
        bands[0] += bands.pop()
    return bands

def find_cuts(sequence, enzymes, linear=True):
    batch = Restriction.RestrictionBatch(enzymes)
    return [
        cut
        for cuts in batch.search(sequence, linear=linear).values()
        for cut in cuts
    ]

def predict_digestion_bands(sequence, enzymes, linear=True):
    """Return the band sizes from digestion by all enzymes at once.

    Returns a list of bands sizes sorted from smallest to largest

    Parameters
    ----------
    sequence
      Sequence to be digested. Either a string ("ATGC...") or
      a BioPython `Seq`

    enzymes
      list of all enzymes placed at the same time in the digestion
      mix e.g. `["EcoRI", "BamHI"]`

    linear
      True if the DNA fragment is linearized, False if it is circular

    """
    if not isinstance(sequence, Seq):
        sequence = Seq(sequence)
    cut_sites = find_cuts(sequence, enzymes, linear=linear)
    bands = _compute_bands_from_cuts(cut_sites, len(sequence), linear=linear)
    return sorted(bands)


def predict_sequence_digestions(sequence, enzymes, linear=True,
                                max_enzymes_per_digestion=1,
                                bands_to_migration=None):
    """Return a dict giving bands sizes pattern for all possible digestions.

    The digestions, double-digestions, etc. are listed and for each the
    sequence band sizes are computed.

    The result if of the form ``{digestion: {'cuts': [], 'bands': []}}``
    Where ``digestion`` is a tuple of enzyme names e.g. ``('EcoRI', 'XbaI')``,
    'cuts' is a list of cuts locations, 'bands' is a list of bands sizes

    Parameters
    ----------
    sequence
      The sequence to be digested

    enzymes
      List of all enzymes to be considered

    max_enzymes_per_digestion
      Maximum number of enzymes allowed in one digestion

    bands_to_migration
      Function associating a migration distance to a band size. If provided,
      each digestion will have a ``'migration'`` field (list of migration
      distances) in addition to 'cuts' and 'bands'.

    """
    restriction_batch = Restriction.RestrictionBatch(enzymes)
    cuts_dict = restriction_batch.search(Seq(sequence))

    def get_cuts(enzyme_name):
        return {"cuts": cuts_dict[Restriction.__dict__[enzyme_name]]}

    def _merge_digestions(digestion1, digestion2, sequence_length, linear):
        """Merge and sort the cuts from two different digestions."""
        all_cuts = sorted(list(set(digestion1["cuts"] + digestion2["cuts"])))
        return {
            "cuts": all_cuts,
            "bands": _compute_bands_from_cuts(
                cuts=all_cuts,
                sequence_length=sequence_length,
                linear=linear
            )
        }
    empty_digestion = ((), {"cuts": [], "bands": [len(sequence)]})
    digestions_dict = OrderedDict([empty_digestion])
    for n_enzymes in range(max_enzymes_per_digestion):
        sub_enzymes = [enzs for enzs in digestions_dict.keys()
                       if len(enzs) == n_enzymes]
        for enzyme in enzymes:
            sub_sub_enzymes = [enzs for enzs in sub_enzymes
                               if enzyme not in enzs]
            for enzs in sub_sub_enzymes:
                digestion = tuple(sorted(enzs + (enzyme,)))
                if digestion not in digestions_dict:
                    no_enzyme_band = len(get_cuts(enzyme)['cuts']) == 0
                    no_enzs_band = len(digestions_dict[enzs]['cuts']) == 0
                    one_no_bands = no_enzs_band or no_enzyme_band
                    if ((enzyme,) in digestions_dict) and one_no_bands:
                        if no_enzyme_band:
                            digestions_dict[digestion] = digestions_dict[enzs]
                        elif no_enzs_band:
                            dig = (enzyme,)
                            digestions_dict[digestion] = digestions_dict[dig]
                    else:
                        digestions_dict[digestion] = _merge_digestions(
                            digestion1=get_cuts(enzyme),
                            digestion2=digestions_dict[enzs],
                            sequence_length=len(sequence),
                            linear=linear
                        )
                        if bands_to_migration is not None:
                            bands = digestions_dict[digestion]["bands"]
                            migration = bands_to_migration(bands)
                            digestions_dict[digestion]["migration"] = migration
                        digestions_dict[digestion]['same_as'] = digestion
    digestions_dict.pop(())

    # Reordering the dictionnary makes the computation of scores faster
    # using 'same_as' to avoid recomputing scores involving similar patterns
    digestions_dict = OrderedDict(
        sorted(digestions_dict.items(),
               key=lambda item: (len(item[0]), len(item[1]['cuts'])))
    )
    return digestions_dict



"""Collection of useful methods and solvers for BandWitch."""

from Bio import Restriction
from .data.enzymes_infos import enzymes_infos
from Bio.Seq import Seq
from Bio import SeqIO
import numpy as np

class NoSolutionError(Exception):
    """Specific error for the case where no enzyme set can cut it. Ah ah."""
    pass

def list_common_enzymes(site_length=(6,), opt_temp=(37,), min_suppliers=1,
                        uniquify_sites=True,
                        avoid_methylations=('Dam', 'Dcm')):
    result = [
    str(e)
        for e in Restriction.AllEnzymes
        if str(e) in enzymes_infos
        and (len(e.site) in site_length)
        and ((opt_temp is None) or (e.opt_temp == 37))
        and enzymes_infos[str(e)]['suppliers'] >= min_suppliers
        and not any([enzymes_infos[str(e)][meth]
                     for meth in avoid_methylations])
    ]
    if uniquify_sites:
        sites_dict = {}
        for e in result:
            enzyme = Restriction.__dict__[e]
            sites_dict[enzyme.site] = e
        result = list(sites_dict.values())
    return sorted(result)

def load_genbank(filename, linear=True, name="unnamed"):
    """Load a genbank from a file."""
    record = SeqIO.read(filename, "genbank")
    record.linear = linear
    record.id = name
    record.name = name.replace(" ", "_")[:20]
    return record

def digestions_list_to_string(digestions):
    """Return a nicely formatted string of a list of 'digestion'
    (where a digestion is a list of enzyme names)"""
    return ", ".join([" + ".join([e for e in d]) for d in digestions])

def max_min_distance(values_1, values_2, zone=None):
    """Return the maximum distance between one value in one set and its
    nearest neighbor in the other set.

    A large max-min-distance means that the sets values_1 and values_2 are
    visually different, i.e. that one value in one of these sets has no close
    equivalent in the other set.

    values_1, values2
      Two lists of values

    zone
      The max distance will be computed only on couples (v1, v2) where v1 is
      in the zone. Allows to remove border effects.
    """
    values_1, values_2 = np.array(values_1), np.array(values_2)
    all_distances = abs(values_1.reshape((len(values_1), 1)) - values_2)
    if zone is not None:
        mini, maxi = zone
        m1 = all_distances.min(axis=0)[(mini <= values_2) & (values_2 <= maxi)]
        m1 = m1.max() if len(m1) else 0
        m2 = all_distances.min(axis=1)[(mini <= values_1) & (values_1 <= maxi)]
        m2 = m2.max() if len(m2) else 0
        return max(m1, m2)

    return max([all_distances.min(axis=i).max() for i in (0, 1)])


def band_patterns_discrepancy(bands1, bands2, ladder, relative_tolerance=0.1,
                              reference_and_gel=False, zone=None):
    """Return a discrepancy indicating whether the band patterns are
    perfectly matching (0) or dissimilar (>=1) or in-between (0-1).

    Similarity is determined by the fact that the min-max distance between
    the two migration patterns is below some threshold. The threshold is some
    proportion of the ladder's migration span.

    If mode ``reference_and_gel`` is set to True, then ``bands1`` is considered
    a "truth" (= expected bands pattern) and the function will automatically
    return False if the observed pattern on gel (bands2) has more bands than
    bands1.

    Parameters
    ----------

    bands1
      A list of bands sizes forming the first pattern

    bands2
      A list of bands sizes forming the second pattern

    ladder
      A Ladder object used to compute migration distances.

    relative tolerance
      Proportion of the ladder's migration span (between lowest and highest
      migration distances) that is the threshold for the min-max distance.

    reference_and_gel
      Set to True if ``bands1`` is a reference (= expected bands) and
      ``bands2`` is the observed bands.

    zone
    """
    if bands1 == bands2 == []:
        return 0
    elif min(len(bands1), len(bands2)) == 0:
        return 2.0
    if reference_and_gel and (len(bands2) > len(bands1)):
        return 2.0

    m1, m2 = (ladder.dna_size_to_migration(np.array(b))
              for b in (bands1, bands2))
    mini, maxi = ladder.migration_distances_span
    tolerance = relative_tolerance * (maxi - mini)
    zone = [ladder.dna_size_to_migration(b) for b in zone][::-1]
    return 1.0 * max_min_distance(m1, m2, zone=zone) / tolerance

def updated_dict(dic1, dic2):
    """Return dic1 updated with dic2 if dic2 is not None.

    Example
    -------

    >>> my_dict = updated_dict({"size": 7, "color": "r"}, some_defaults_dict)
    """
    if dic2 is not None:
        dic1.update(dic2)
    return dic1
